import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from aporia.sdk.client import Client
from aporia.sdk.datasets import DatasetType


class TimeRange(BaseModel):
    start: datetime.datetime
    end: datetime.datetime  # "2023-05-12T00:00:00.000000+00:00"

    def serialize(self):
        return {
            "start": self.start.isoformat(timespec="microseconds") + "+00:00",
            "end": self.end.isoformat(timespec="microseconds") + "+00:00",
        }


class MetricSegment(BaseModel):
    id: str
    value: str  # TODO: Test with numeric auto segments

    def serialize(self) -> Dict:
        return {"id": self.id, "value": self.value}


class MetricDataset(BaseModel):
    dataset_type: DatasetType = DatasetType.SERVING
    time_range: Optional[TimeRange] = None
    model_version: Optional[str] = None
    segment: Optional[MetricSegment] = None

    def serialize(self) -> Dict:
        return {
            "dataset_type": self.dataset_type.value,
            "time_range": self.time_range.serialize() if self.time_range is not None else None,
            "model_version": self.model_version,
            "segment": self.segment.serialize() if self.segment is not None else None,
        }


class MetricResponse(BaseModel):
    id: str
    value: Any
    error: Optional[Union[Dict, List, str]]
    segment: Optional[MetricSegment]


class QueryResponse(BaseModel):
    metrics: List[MetricResponse]
    time_range: TimeRange


class MetricParameters(BaseModel):
    dataset: MetricDataset
    # Metric identifier
    name: str
    # Parameters
    column: Optional[str] = None
    k: Optional[int] = None
    threshold: Optional[float] = None
    custom_metric_id: Optional[str] = None
    baseline: Optional[MetricDataset] = None


class MetricError(Exception):
    def __init__(self, error, parameters: MetricParameters):
        super().__init__(error)
        self.parameters = parameters


class AporiaMetrics:
    def __init__(
        self,
        token: str,
        account_name: str,
        base_url: str = "https://platform.aporia.com",
        workspace_name: str = "default-workspace",
    ) -> Dict:
        self.client = Client(
            base_url=f"{base_url}/api/v1/{account_name}/{workspace_name}", token=token
        )

    def query(
        self,
        # Dataset identifiers
        model_id: str,
        dataset: MetricDataset,
        # Metric identifier
        metric_name: str,
        baseline: Optional[MetricDataset] = None,
        # Parameters
        column: Optional[str] = None,
        k: Optional[int] = None,
        threshold: Optional[float] = None,
        custom_metric_id: Optional[str] = None,
    ):
        metric_parameters = {}
        if column is not None:
            metric_parameters["column"] = column
        if k is not None:
            metric_parameters["k"] = k
        if threshold is not None:
            metric_parameters["threshold"] = threshold
        if custom_metric_id is not None:
            metric_parameters["id"] = custom_metric_id

        metric_datasets = {"data": dataset.serialize()}
        if baseline is not None:
            metric_datasets["baseline"] = baseline.serialize()

        response = self.client.send_request(
            "/query",
            method="POST",
            data={
                "model_id": model_id,
                "metrics": [
                    {
                        "id": "0",
                        "metric": metric_name,
                        "parameters": metric_parameters,
                        "datasets": metric_datasets,
                    }
                ],
            },
            url_search_replace=("/api/v1/", "/v1/metrics-reducer/"),
        )
        self.client.assert_response(response)

        result = [QueryResponse(**entry) for entry in response.json()]

        if result[0].metrics[0].error is not None:
            raise Exception(f"Error occured: {result[0].metrics[0].error}")
        return result[0].metrics[0].value

    def query_batch(self, model_id: str, metrics: List[MetricParameters]) -> List:
        metric_requests = []
        for i, metric in enumerate(metrics):
            metric_parameters = {}
            if metric.column is not None:
                metric_parameters["column"] = metric.column
            if metric.k is not None:
                metric_parameters["k"] = metric.k
            if metric.threshold is not None:
                metric_parameters["threshold"] = metric.threshold
            if metric.custom_metric_id is not None:
                metric_parameters["id"] = metric.custom_metric_id

            metric_datasets = {"data": metric.dataset.serialize()}
            if metric.baseline is not None:
                metric_datasets["baseline"] = metric.baseline.serialize()
            metric_requests.append(
                {
                    "id": str(i),
                    "metric": metric.name,
                    "parameters": metric_parameters,
                    "datasets": metric_datasets,
                }
            )

        response = self.client.send_request(
            "/query",
            method="POST",
            data={
                "model_id": model_id,
                "metrics": metric_requests,
            },
            url_search_replace=("/api/v1/", "/v1/metrics-reducer/"),
        )
        self.client.assert_response(response)

        result = [QueryResponse(**entry) for entry in response.json()]

        # Restore order, because different timeframes may change order
        metric_results = {}
        for entry in result:
            for metric in entry.metrics:
                metric_index = int(metric.id)
                if metric.error is not None:
                    metric_results[metric_index] = MetricError(
                        error=metric.error, parameters=metrics[metric_index]
                    )
                else:
                    metric_results[metric_index] = metric.value

        return [metric_results[i] for i in range(len(metric_results))]



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.data_sources import DataSource


class DatasetType(Enum):
    SERVING = "serving"
    TRAINING = "training"


class FieldType(Enum):
    BOOLEAN = "boolean"
    NUMERIC = "numeric"
    DATETIME = "datetime"
    CATEGORICAL = "categorical"
    TEXT = "text"
    ARRAY = "array"
    EMBEDDING = "embedding"
    IMAGE_URL = "image_url"


class FieldSchema(BaseModel):
    name: str
    type: FieldType
    properties: Optional[Dict[str, Any]] = None


class SortOrder(Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class DatasetSchema(BaseModel):
    # Null id and timestamp allowed for training
    id_column: Optional[str]
    timestamp_column: Optional[str]
    order_by_column: Optional[str] = None
    group_by_column: Optional[str] = None
    group_order_sort: Optional[SortOrder] = None
    raw_inputs: List[FieldSchema]
    features: List[FieldSchema]
    predictions: List[FieldSchema]
    actuals: List[FieldSchema]


class Dataset(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]

    @classmethod
    def get_all(cls, client: Client, model_version_id: Optional[str] = None) -> List["Dataset"]:
        response = client.send_request(
            f"/datasets{'' if model_version_id is None else f'?model_version_id={model_version_id}'}",
            "GET",
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        model_id: str,
        version_id: str,
        dataset_type: DatasetType,
        connection_data: Dict[str, Any],
        schema: DatasetSchema,
        data_source: Optional[DataSource] = None,
        data_source_id: Optional[str] = None,
    ) -> "Dataset":
        dataset_type = DatasetType(dataset_type)

        group_keys = {}

        if schema.group_by_column is not None:
            group_keys["group_by_column"] = schema.group_by_column
        if schema.order_by_column is not None:
            group_keys["order_by_column"] = schema.order_by_column
        if schema.group_order_sort is not None:
            group_keys["group_order_sort"] = schema.group_order_sort.value

        response = client.send_request(
            "/datasets",
            "POST",
            {
                "stage": dataset_type.value,
                "model_version_id": version_id,
                "model_id": model_id,
                "data_source_id": data_source.id if data_source_id is None else data_source_id,
                "config": connection_data,
                "schema": {
                    "id_column": schema.id_column,
                    "timestamp_column": schema.timestamp_column,
                    **group_keys,
                    "raw_inputs": [
                        {"name": raw_input.name, "type": raw_input.type.value}
                        for raw_input in schema.raw_inputs
                    ],
                    "features": [
                        {"name": feature.name, "type": feature.type.value}
                        for feature in schema.features
                    ],
                    "predictions": [
                        {"name": prediction.name, "type": prediction.type.value}
                        for prediction in schema.predictions
                    ],
                    "actuals": [
                        {
                            "name": actual.name,
                            "type": actual.type.value,
                            "properties": actual.properties,
                        }
                        for actual in schema.actuals
                    ],
                },
            },
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Dataset":
        response = client.send_request(f"/datasets/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(
        self,
        # dataset_type: DatasetType | None = None,
        connection_data: Optional[Dict[str, Any]] = None,
        schema: Optional[DatasetSchema] = None,
        data_source: Optional[DataSource] = None,
        data_source_id: Optional[str] = None,
        **kwargs,
    ):
        # dataset_type = DatasetType(dataset_type)

        args = {}
        # if dataset_type is not None:
        #     args["stage"] = dataset_type.value
        if data_source_id is not None or data_source is not None:
            args["data_source_id"] = data_source.id if data_source_id is None else data_source_id
        if connection_data is not None:
            args["config"] = connection_data
        # if schema is not None:
        #     args["schema"] = {
        #         "id_column": schema.id_column,
        #         "timestamp_column": schema.timestamp_column,
        #         "raw_inputs": [
        #             {"name": raw_input.name, "type": raw_input.type.value}
        #             for raw_input in schema.raw_inputs
        #         ],
        #         "features": [
        #             {"name": feature.name, "type": feature.type.value}
        #             for feature in schema.features
        #         ],
        #         "predictions": [
        #             {"name": prediction.name, "type": prediction.type.value}
        #             for prediction in schema.predictions
        #         ],
        #         "actuals": [
        #             {
        #                 "name": actual.name,
        #                 "type": actual.type.value,
        #                 "properties": actual.properties,
        #             }
        #             for actual in schema.actuals
        #         ],
        #     }
        if schema is not None:
            if schema.id_column is not None:
                args["id_column"] = schema.id_column
            if schema.timestamp_column is not None:
                args["timestamp_column"] = schema.timestamp_column

        # response = self.client.send_request(
        #     f"/datasets/{self.id}",
        #     "PUT",
        #     args,
        # )
        response = self.client.send_request(
            f"/datasets/{self.id}/change-query",
            "PUT",
            args,
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)
        # self.__update_members(response.json())
        data = response.json()
        self.raw_data["config"] = data["config"]
        self.raw_data["schema"]["id_column"] = data["id_column"]
        self.raw_data["schema"]["timestamp_column"] = data["timestamp_column"]
        self.raw_data["data_source_id"] = data["data_source_id"]

    def delete(self):
        raise NotImplementedError()

    @staticmethod
    def delete_by_id(client: Client, id: str):
        raise NotImplementedError()



from typing import Dict, Optional, Tuple

import requests


class Client:
    def __init__(self, base_url: str, token: str, debug: bool = False):
        self.base_url = base_url
        self.token = token
        self.debug = debug

    def send_request(
        self,
        url: str,
        method: str,
        data: Optional[Dict] = None,
        url_override: Optional[str] = None,
        url_search_replace: Optional[Tuple[str, str]] = None,
    ):
        if url_override is not None:
            target_url = url_override
        else:
            target_url = self.base_url + url

        if url_search_replace is not None:
            target_url = target_url.replace(url_search_replace[0], url_search_replace[1])

        if self.debug:
            print(f"{method} {target_url}, ({data})", end=" -> ")

        response = requests.request(
            method=method,
            url=target_url,
            json=data,
            headers={"Authorization": f"Bearer {self.token}"},
        )

        if self.debug:
            print(f"{response.status_code}")

        return response

    def assert_response(self, response: requests.Response):
        if not response.ok:
            try:
                body = response.json()
            except Exception:
                body = response.content
            if self.debug:
                print(f"Request failed with status {response.status_code}, body: {body}")
            raise RuntimeError(f"Request failed with status {response.status_code}, body: {body}")



from typing import Dict, List, Optional

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client


class CustomMetric(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.id = data["id"]
        self.name = data["name"]
        self.raw_data = data

    @classmethod
    def get_all(cls, client: Client, model_id: Optional[str] = None) -> List["CustomMetric"]:
        response = client.send_request(
            f"/custom-metrics{'' if model_id is None else f'?model_id={model_id}'}", "GET"
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        model_id: str,
        code: str,
    ) -> "CustomMetric":
        response = client.send_request(
            "/custom-metrics",
            "POST",
            {"name": name, "model_id": model_id, "code": code},
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "CustomMetric":
        response = client.send_request(f"/custom-metrics/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(self, **kwargs):
        response = self.client.send_request(
            f"/custom-metrics/{self.id}",
            "PUT",
            {"syntax": "aporia", **kwargs},
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/custom-metrics/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/custom-metrics/{id}", "DELETE")
        client.assert_response(response)



from typing import Any, Dict, Optional, Tuple, Union

from aporia.as_code.resources.base import BaseResource, CompareStatus
from aporia.sdk.client import Client
from aporia.sdk.data_sources import DataSource as _DataSource
from aporia.sdk.data_sources import DataSourceType


class DataSource(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        type: Union[DataSourceType, str],
        connection_data: Dict[str, Any],
        name: Optional[str] = None,
    ):
        self.dependants = []
        self.name = resource_name
        if name is None:
            name = resource_name

        self._args = {
            "name": name,
            "data_source_type": DataSourceType(type),
            "connection_data": connection_data,
        }

    def compare(self, resource_data: Dict) -> CompareStatus:
        if all(
            [
                self._args["name"] == resource_data["name"],
                self._args["data_source_type"].value == resource_data["type"],
                # TODO: Data source configuration isn't getable
                # self._args["connection_data"] == resource_data["connection"],
            ]
        ):
            return CompareStatus.SAME
        return CompareStatus.MISMATCHED

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        data_source = _DataSource.create(client=client, **self._args)
        return data_source.id, data_source.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _DataSource.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        data_source = _DataSource.read(client=client, id=id)
        data_source.update(**self._args)
        return data_source.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        _DataSource.delete_by_id(client=client, id=id)

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        if self._args["name"] != resource_data["name"]:
            diffs["name"] = (resource_data["name"], self._args["name"])
        if self._args["data_source_type"].value != resource_data["type"]:
            diffs["type"] = (resource_data["type"], self._args["data_source_type"].value)
        return diffs



from typing import Any, Dict, Optional, Tuple

from aporia.as_code.resources.base import BaseResource, CompareStatus
from aporia.sdk.client import Client
from aporia.sdk.custom_metrics import CustomMetric as _CustomMetric


class CustomMetric(BaseResource):
    def __init__(self, resource_name: str, /, *, code: str, name: Optional[str] = None):
        self.name = resource_name
        self.dependants = []
        if name is None:
            name = resource_name

        self._args = {"name": name, "code": code}

    def compare(self, resource_data: Dict) -> CompareStatus:
        if all([self._args[k] == resource_data[k] for k in self._args.keys()]):
            return CompareStatus.SAME
        return CompareStatus.UPDATEABLE

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        custom_metric = _CustomMetric.create(client=client, **self._args)
        return custom_metric.id, custom_metric.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _CustomMetric.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        custom_metric = _CustomMetric.read(client=client, id=id)
        custom_metric.update(**self._args)
        return custom_metric.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        _CustomMetric.delete_by_id(client=client, id=id)

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        for k in self._args.keys():
            if self._args[k] != resource_data[k]:
                diffs[k] = (resource_data[k], self._args[k])
        return diffs



from typing import Any, Dict, Optional, Tuple

from aporia.as_code.resources.base import BaseResource, CompareStatus, NonDeletableResourceException
from aporia.as_code.resources.dataset import Dataset
from aporia.sdk.client import Client
from aporia.sdk.versions import Version as _Version


class Version(BaseResource):
    def __init__(
        self,
        resource_name: str,
        /,
        *,
        name: Optional[str] = None,
        serving: Optional[Dataset] = None,
        training: Optional[Dataset] = None,
    ):
        self.name = resource_name
        self.dependants = []
        self.sub_resources = []
        if name is None:
            name = resource_name

        self._args = {"name": name}

        def apply_args(data, dataset):
            dataset.setarg("version_id", data["id"])
            dataset.setarg("model_id", data["model_id"])

        if serving is not None:
            self.sub_resources.append((serving, apply_args))
        if training is not None:
            self.sub_resources.append((training, apply_args))

    def compare(self, resource_data: Dict) -> CompareStatus:
        if all([self._args[k] == resource_data[k] for k in self._args.keys()]):
            return CompareStatus.SAME
        return CompareStatus.UPDATEABLE

    def setarg(self, arg_name: str, arg_value: Any):
        self._args[arg_name] = arg_value

    def create(self, client: Client) -> Tuple[str, Dict]:
        version = _Version.create(client=client, **self._args)
        return version.id, version.raw_data

    def read(self, client: Client, id: str) -> Dict:
        return _Version.read(client=client, id=id).raw_data

    def update(self, client: Client, id: str) -> Dict:
        version = _Version.read(client=client, id=id)
        version.update(**self._args)
        return version.raw_data

    @classmethod
    def delete(cls, client: Client, id: str):
        raise NonDeletableResourceException()

    def get_diff(self, resource_data: Dict) -> Dict:
        diffs = {}
        for k in self._args.keys():
            if self._args[k] != resource_data[k]:
                diffs[k] = (resource_data[k], self._args[k])
        return diffs



from enum import Enum
from typing import Any, Dict, Type

from pydantic import BaseModel, validator


class WidgetType(Enum):
    METRIC = "metric"
    TIMESERIES = "timeseries"
    DISTRIBUTION = "distribution"
    TABLE = "datahealthtable"
    TEXT = "text"


class BaseWidgetData(BaseModel):
    pass


class MetricWidgetData(BaseWidgetData):
    pass


class TimeseriesWidgetData(BaseWidgetData):
    pass


class TableWidgetData(BaseWidgetData):
    pass


class DistributionWidgetData(BaseWidgetData):
    pass


class TextWidgetData(BaseWidgetData):
    text: str


WIDGET_TYPE_TO_WIDGET_DATA_CLASS: Dict[WidgetType, Type[BaseWidgetData]] = {
    WidgetType.METRIC: MetricWidgetData,
    WidgetType.TIMESERIES: TimeseriesWidgetData,
    WidgetType.DISTRIBUTION: DistributionWidgetData,
    WidgetType.TABLE: TableWidgetData,
    WidgetType.TEXT: TextWidgetData,
}


class Widget(BaseModel):
    loc_x: int
    loc_y: int
    type: WidgetType
    data: BaseWidgetData

    @validator("data", pre=True)
    @classmethod
    def _parse_data(cls, value: Any, values: Dict[str, Any]) -> BaseWidgetData:
        """Validates data matches the widget type.

        Args:
            value: data dict to validate
            values: other attributes of Widget (type)

        Returns:
            BaseWidgetData: WidgetData object
        """
        # Parsing `data` requires `type` to be parsed
        if "type" not in values:
            return value

        if values["type"] not in WIDGET_TYPE_TO_WIDGET_DATA_CLASS:
            raise ValueError(f"Unsupported widget type {values['type']}")

        if not isinstance(value, dict):
            raise ValueError("expected a dict containing widget data")

        parameters_cls = WIDGET_TYPE_TO_WIDGET_DATA_CLASS[values["type"]]
        return parameters_cls(**value)



from aporia.as_code.resources.custom_metrics import CustomMetric
from aporia.as_code.resources.data_source import DataSource
from aporia.as_code.resources.dataset import Dataset
from aporia.as_code.resources.model import Model
from aporia.as_code.resources.monitor import Monitor
from aporia.as_code.resources.segment import Segment
from aporia.as_code.resources.version import Version
from aporia.as_code.stack import Stack

# Consts
from aporia.sdk.data_sources import DataSourceType
from aporia.sdk.datasets import DatasetType, SortOrder
from aporia.sdk.models import ModelAggregationPeriod, ModelColor, ModelIcon, ModelType
from aporia.sdk.monitors import (
    AverageMethod,
    BaselineConfiguration,
    DetectionMethod,
    FocalConfiguration,
    MessagingIntegrationType,
    MetricType,
    MonitorType,
    PeriodType,
    SEGMENT_ID_ALL_DATA,
    SegmentIdentification,
    Severity,
    SourceType,
    ThresholdConfiguration,
    TimePeriod,
)



from typing import Any, Dict, List, Optional

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.data_sources import DataSource
from aporia.sdk.datasets import Dataset, DatasetSchema, DatasetType


class Version(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.model_id = data["model_id"]

    @classmethod
    def get_all(cls, client: Client, model_id: Optional[str] = None) -> List["Version"]:
        response = client.send_request(
            f"/model-versions{'' if model_id is None else f'?model_id={model_id}'}", "GET"
        )

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(cls, client: Client, name: str, model_id: str) -> "Version":
        existing_versions = cls.get_all(client=client, model_id=model_id)

        for version in existing_versions:
            if version.name == name:
                print(f"Version {name} already exists! Using it. Check datasets")
                return version

        response = client.send_request(
            "/model-versions",
            "POST",
            {"name": name, "model_id": model_id},
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    def create_dataset(
        self,
        dataset_type: DatasetType,
        data_source: DataSource,
        connection_data: Dict[str, Any],
        schema: DatasetSchema,
    ) -> Dataset:
        """Creates a new dataset."""
        dataset = Dataset.create(
            client=self.client,
            model_id=self.model_id,
            version_id=self.id,
            dataset_type=dataset_type,
            data_source=data_source,
            connection_data=connection_data,
            schema=schema,
        )
        return dataset

    @classmethod
    def read(cls, client: Client, id: str) -> "Version":
        response = client.send_request(f"/model-versions/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(self, **kwargs):
        response = self.client.send_request(
            f"/model-versions/{self.id}",
            "PUT",
            {"name": kwargs["name"]},
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        raise NotImplementedError()

    @staticmethod
    def delete_by_id(client: Client, id: str):
        raise NotImplementedError()

    def get_datasets(self) -> List[Dataset]:
        datasets = Dataset.get_all(client=self.client, model_version_id=self.id)
        return datasets



from enum import Enum
from typing import Dict, List, Optional, Union

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.custom_metrics import CustomMetric
from aporia.sdk.dashboards import Widget
from aporia.sdk.monitors import Monitor, MonitorConfiguration, MonitorType
from aporia.sdk.segments import Segment
from aporia.sdk.versions import Version


class ModelType(Enum):
    REGRESSION = "regression"
    BINARY = "binary"
    MULTICLASS = "multiclass"
    MULTILABEL = "multi-label"
    RANKING = "ranking"


class ModelIcon(Enum):
    GENERAL = "general"
    CHURN_RETENTION = "churn-and-retention"
    CONVERSION_PREDICT = "conversion-predict"
    ANOMALY = "anomaly"
    DYNAMIC_PRICING = "dynamic-pricing"
    EMAIL_FILTERING = "email-filtering"
    DEMAND_FORECASTING = "demand-forecasting"
    LTV = "ltv"
    PERSONALIZATION = "personalization"
    FRAUD_DETECTION = "fraud-detection"
    CREDIT_RISK = "credit-risk"
    RECOMMENDATIONS = "recommendations"


class ModelColor(Enum):
    BLUE = "#3564C9"
    TURQUOISE = "#60DFE8"
    GREEN = "#A0D468"
    LIGHT_TEAL = "#39DAA3"
    PINK = "#EC87C0"
    PURPLE = "#AF72FD"
    GOLD = "#FFCE54"
    ORANGE = "#FC6E51"


class ModelAggregationPeriod(Enum):
    DAILY = "daily"
    HOURLY = "hourly"


class NoOwner:
    pass


class Model(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.type = ModelType(data["type"])
        self.description = data["description"]
        try:
            self.color = ModelColor(data["color"])
        except Exception:
            self.color = data["color"]
        self.icon = ModelIcon(data["icon"])
        self.owner = data["owner"]

    @classmethod
    def get_all(cls, client: Client) -> List["Model"]:
        response = client.send_request("/models", "GET")

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        model_type: ModelType,
        description: Optional[str] = None,
        icon: Optional[ModelIcon] = None,
        color: Optional[ModelColor] = None,
        owner: Optional[Union[str, NoOwner]] = NoOwner(),
        aggregation_period: Optional[ModelAggregationPeriod] = None,
    ) -> "Model":
        """Creates a new model in Aporia, and returns a new model descriptor."""
        model_type = ModelType(model_type)

        creation_parameters = {"name": name, "type": model_type.value}

        if description is not None:
            creation_parameters["description"] = description
        if icon is not None:
            creation_parameters["icon"] = icon.value
        if color is not None:
            creation_parameters["color"] = color.value
        if not isinstance(owner, NoOwner):
            creation_parameters["owner"] = owner
        if aggregation_period is not None:
            creation_parameters["aggregation_period"] = aggregation_period.value

        response = client.send_request(
            "/models",
            "POST",
            creation_parameters,
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "Model":
        response = client.send_request(f"/models/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(self, **kwargs):
        args = {}
        for k, v in kwargs.items():
            if isinstance(v, Enum):
                args[k] = v.value
            else:
                args[k] = v
        response = self.client.send_request(f"/models/{self.id}", "PUT", args)
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(f"/models/{self.id}", "DELETE")
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(f"/models/{id}", "DELETE")
        client.assert_response(response)

    def create_model_version(self, name: str) -> Version:
        """Creates a new model version."""
        version = Version.create(client=self.client, name=name, model_id=self.id)
        return version

    def create_custom_metric(self, name: str, code: str) -> CustomMetric:
        metric = CustomMetric.create(client=self.client, name=name, model_id=self.id, code=code)
        return metric

    def create_segment(
        self,
        name: str,
        field_name: Optional[str] = None,
        values: Optional[Union[List[str], List[Union[float, int]]]] = None,
        term: Optional[str] = None,
    ) -> Segment:
        segment = Segment.create(
            client=self.client,
            name=name,
            model_id=self.id,
            field_name=field_name,
            values=values,
            term=term,
        )
        return segment

    def create_monitor(
        self,
        name: str,
        monitor_type: MonitorType,
        configuration: MonitorConfiguration,
        scheduling: Optional[str] = None,
        comment: Optional[str] = None,
        creator: Optional[str] = None,
        is_active: bool = True,
    ) -> Monitor:
        monitor = Monitor.create(
            client=self.client,
            model_id=self.id,
            name=name,
            monitor_type=monitor_type,
            scheduling=scheduling,
            configuration=configuration,
            comment=comment,
            creator=creator,
            is_active=is_active,
        )
        return monitor

    def create_dashboard(self, widgets: List[Widget]):
        pass

    def get_versions(self) -> List[Version]:
        versions = Version.get_all(client=self.client, model_id=self.id)
        return versions

    def get_segments(self) -> List[Segment]:
        segments = Segment.get_all(client=self.client, model_id=self.id)
        return segments

    def get_custom_metrics(self) -> List[CustomMetric]:
        custom_metrics = CustomMetric.get_all(client=self.client, model_id=self.id)
        return custom_metrics

    def get_monitors(self) -> List[Monitor]:
        monitors = Monitor.get_all(client=self.client, model_id=self.id)
        return monitors

    def _get_calculation_status(self) -> str:
        response = self.client.send_request(
            "/default_metrics/get_status",
            "POST",
            [{"model_id": self.id}],
            url_search_replace=("/api/v1/", "/v1/metrics-agency/"),
        )

        self.client.assert_response(response)

        return response.json()["statuses"][0]



from abc import ABC, abstractclassmethod, abstractmethod, abstractstaticmethod
from typing import List

from aporia.sdk.client import Client


class BaseAporiaResource(ABC):
    @abstractclassmethod
    def create(cls, client: Client, name: str, *args, **kwargs) -> "BaseAporiaResource":
        ...

    @abstractclassmethod
    def read(cls, client: Client, id: str) -> "BaseAporiaResource":
        ...

    @abstractmethod
    def update(self):
        ...

    @abstractmethod
    def delete(self):
        ...

    @abstractclassmethod
    def get_all(cls, client: Client, **kwargs) -> List["BaseAporiaResource"]:
        ...

    @abstractstaticmethod
    def delete_by_id(cls, client: Client, id: str):
        ...



from typing import Any, Dict, List, Optional, Union

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client
from aporia.sdk.data_sources import DataSource, DataSourceType
from aporia.sdk.metrics import AporiaMetrics, MetricParameters
from aporia.sdk.models import Model, ModelColor, ModelIcon, ModelType, NoOwner


class Aporia:
    def __init__(
        self,
        token: str,
        account_name: str,
        workspace_name: str = "default-workspace",
        base_url: str = "https://platform.aporia.com",
        debug: bool = False,
        delete_on_failure: bool = False,
    ):
        self._assert_workspace(
            base_url=base_url, token=token, account=account_name, workspace=workspace_name
        )
        self.client = Client(
            base_url=f"{base_url}/api/v1/{account_name}/{workspace_name}", token=token, debug=debug
        )
        self.metrics_client = AporiaMetrics(
            token=token, account_name=account_name, workspace_name=workspace_name, base_url=base_url
        )
        self.delete_on_failure = delete_on_failure
        self._created_resources: List[BaseAporiaResource] = []

    def create_model(
        self,
        name: str,
        model_type: ModelType,
        description: Optional[str] = None,
        icon: Optional[ModelIcon] = None,
        color: Optional[ModelColor] = None,
        owner: Optional[Union[str, NoOwner]] = NoOwner(),
    ) -> Model:
        model = Model.create(
            client=self.client,
            name=name,
            model_type=model_type,
            description=description,
            icon=icon,
            color=color,
            owner=owner,
        )
        self._created_resources.append(model)
        return model

    def create_data_source(
        self, name: str, data_source_type: DataSourceType, connection_data: Dict[str, Any]
    ) -> DataSource:
        data_source = DataSource.create(
            client=self.client,
            name=name,
            data_source_type=data_source_type,
            connection_data=connection_data,
        )
        self._created_resources.append(data_source)
        return data_source

    def query_metrics(self, model_id: str, metrics: List[MetricParameters]) -> List:
        return self.metrics_client.query_batch(model_id=model_id, metrics=metrics)

    def get_models(self) -> List[Model]:
        models = Model.get_all(client=self.client)
        return models

    def get_data_sources(self) -> List[DataSource]:
        data_sources = DataSource.get_all(client=self.client)
        return data_sources

    def delete(self):
        for resource in self._created_resources:
            resource.delete()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            if self.delete_on_failure:
                self.delete()

    def _assert_workspace(self, base_url: str, token: str, account: str, workspace: str):
        client = Client(base_url=base_url, token=token)
        response = client.send_request(
            f"/v1/identity-service/accounts/{account}/workspaces/{workspace}", "GET"
        )
        client.assert_response(response)



from enum import Enum
from typing import Any, Dict, List

from aporia.sdk.base import BaseAporiaResource
from aporia.sdk.client import Client


class DataSourceType(Enum):
    S3 = "s3"
    GLUE = "glue"
    ATHENA = "athena"
    BIGQUERY = "bigquery"
    POSTGRES = "postgres"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    DATABRICKS = "databricks"
    AZURE_BLOB_STORAGE = "azure_blob_storage"
    SNOWFLAKE_NATIVE = "snowflake_native"
    GOOGLE_CLOUD_STORAGE = "google_cloud_storage"
    HIVE = "hive"


class DataSource(BaseAporiaResource):
    def __init__(self, client: Client, data: Dict):
        self.client = client
        self.__update_members(data)

    def __update_members(self, data: Dict):
        self.raw_data = data
        self.id = data["id"]
        self.name = data["name"]
        self.type = DataSourceType(data["type"])

    @classmethod
    def get_all(cls, client: Client) -> List["DataSource"]:
        response = client.send_request("/data-sources", "GET")

        client.assert_response(response)

        return [cls(client=client, data=entry) for entry in response.json()]

    @classmethod
    def create(
        cls,
        client: Client,
        name: str,
        data_source_type: DataSourceType,
        connection_data: Dict[str, Any],
    ) -> "DataSource":
        existing_data_sources = cls.get_all(client=client)

        for data_source in existing_data_sources:
            if data_source.name == name:
                if data_source.type != data_source_type:
                    raise RuntimeError(
                        f"Data source {name} already exists but is of different type (expected {data_source_type.value}, received {data_source.type.value})"
                    )
                print(f"Data source {name} already exists! Using it. Check configuration")
                return data_source

        connection_data['type'] = data_source_type.value
        response = client.send_request(
            "/data-sources",
            "POST",
            {"name": name, "type": data_source_type.value, "connection": connection_data},
            # TODO: Hack until data-sources are creatable using stable API
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )

        client.assert_response(response)

        return cls(client=client, data=response.json())

    @classmethod
    def read(cls, client: Client, id: str) -> "DataSource":
        response = client.send_request(f"/data-sources/{id}", "GET")
        client.assert_response(response)
        return cls(client=client, data=response.json())

    def update(self, **kwargs):
        args = {}
        for k, v in kwargs.items():
            if isinstance(v, Enum):
                args[k] = v.value
            else:
                args[k] = v
        response = self.client.send_request(f"/data-sources/{self.id}", "PUT", args)
        self.client.assert_response(response)
        self.__update_members(response.json())

    def delete(self):
        response = self.client.send_request(
            f"/data-sources/{self.id}",
            "DELETE",
            # TODO: Hack until data-sources are deleteable using stable API
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        self.client.assert_response(response)

    @staticmethod
    def delete_by_id(client: Client, id: str):
        response = client.send_request(
            f"/data-sources/{id}",
            "DELETE",
            # TODO: Hack until data-sources are deleteable using stable API
            url_search_replace=("/api/v1/", "/v1/crud-service/"),
        )
        client.assert_response(response)
